/*
 * Copyright (c) 2006 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2005 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: osm_sa_multipath_record_ctrl.h 7005 2006-05-09 11:08:59Z halr $
 */


/*
 * Abstract:
 * 	Declaration of osm_mpr_rcv_ctrl_t.
 *	This object represents a controller that receives the IBA
 *	MultiPathRecord attribute from a node.
 *	This object is part of the OpenSM family of objects.
 *
 * Environment:
 * 	Linux User Mode
 *
 */


#ifndef _OSM_MPRCTRL_H_
#define _OSM_MPRCTRL_H_


#include <complib/cl_dispatcher.h>
#include <opensm/osm_base.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_sa_multipath_record.h>

#ifdef __cplusplus
#  define BEGIN_C_DECLS extern "C" {
#  define END_C_DECLS   }
#else /* !__cplusplus */
#  define BEGIN_C_DECLS
#  define END_C_DECLS
#endif /* __cplusplus */

BEGIN_C_DECLS

/****h* OpenSM/MultiPath Record Receive Controller
* NAME
*	MultiPath Record Receive Controller
*
* DESCRIPTION
*	The MultiPath Record Receive Controller object encapsulates
*	the information	needed to receive the MultiPathRecord attribute from a node.
*
*	The MultiPath record Receive Controller object is thread safe.
*
*	This object should be treated as opaque and should be
*	manipulated only through the provided functions.
*
* AUTHOR
*	Hal Rosenstock, Voltaire 
*
*********/
/****s* OpenSM: MultiPath Record Receive Controller/osm_mpr_rcv_ctrl_t
* NAME
*	osm_mpr_rcv_ctrl_t
*
* DESCRIPTION
*	MultiPath Record Receive Controller structure.
*
*	This object should be treated as opaque and should
*	be manipulated only through the provided functions.
*
* SYNOPSIS
*/
typedef struct _osm_mpr_rcv_ctrl
{
	osm_mpr_rcv_t				*p_rcv;
	osm_log_t					*p_log;
	cl_dispatcher_t				*p_disp;
	cl_disp_reg_handle_t		h_disp;

} osm_mpr_rcv_ctrl_t;
/*
* FIELDS
*	p_rcv
*		Pointer to the MultiPath Record Receiver object.
*
*	p_log
*		Pointer to the log object.
*
*	p_disp
*		Pointer to the Dispatcher.
*
*	h_disp
*		Handle returned from dispatcher registration.
*
* SEE ALSO
*	MultiPath Record Receive Controller object
*	MultiPath Record Receiver object
*********/

/****f* OpenSM: MultiPath Record Receive Controller/osm_pr_rcv_ctrl_construct
* NAME
*	osm_mpr_rcv_ctrl_construct
*
* DESCRIPTION
*	This function constructs a MultiPath Record Receive Controller object.
*
* SYNOPSIS
*/
void osm_mpr_rcv_ctrl_construct(
	IN osm_mpr_rcv_ctrl_t* const p_ctrl );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to a MultiPath Record Receive Controller
*		object to construct.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	Allows calling osm_mpr_rcv_ctrl_init, osm_mpr_rcv_ctrl_destroy,
*	and osm_mpr_rcv_ctrl_is_inited.
*
*	Calling osm_mpr_rcv_ctrl_construct is a prerequisite to calling any
*	other method except osm_mpr_rcv_ctrl_init.
*
* SEE ALSO
*	MultiPath Record Receive Controller object, osm_mpr_rcv_ctrl_init,
*	osm_mpr_rcv_ctrl_destroy, osm_mpr_rcv_ctrl_is_inited
*********/

/****f* OpenSM: MultiPath Record Receive Controller/osm_mpr_rcv_ctrl_destroy
* NAME
*	osm_mpr_rcv_ctrl_destroy
*
* DESCRIPTION
*	The osm_mpr_rcv_ctrl_destroy function destroys the object, releasing
*	all resources.
*
* SYNOPSIS
*/
void osm_mpr_rcv_ctrl_destroy(
	IN osm_mpr_rcv_ctrl_t* const p_ctrl );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to the object to destroy.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	Performs any necessary cleanup of the specified
*	MultiPath Record Receive Controller object.
*	Further operations should not be attempted on the destroyed object.
*	This function should only be called after a call to
*	osm_mpr_rcv_ctrl_construct or osm_mpr_rcv_ctrl_init.
*
* SEE ALSO
*	MultiPath Record Receive Controller object, osm_mpr_rcv_ctrl_construct,
*	osm_mpr_rcv_ctrl_init
*********/

/****f* OpenSM: MultiPath Record Receive Controller/osm_mpr_rcv_ctrl_init
* NAME
*	osm_mpr_rcv_ctrl_init
*
* DESCRIPTION
*	The osm_mpr_rcv_ctrl_init function initializes a
*	MultiPath Record Receive Controller object for use.
*
* SYNOPSIS
*/
ib_api_status_t osm_mpr_rcv_ctrl_init(
	IN osm_mpr_rcv_ctrl_t* const p_ctrl,
	IN osm_mpr_rcv_t* const p_rcv,
	IN osm_log_t* const p_log,
	IN cl_dispatcher_t* const p_disp );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to an osm_mpr_rcv_ctrl_t object to initialize.
*
*	p_rcv
*		[in] Pointer to an osm_mpr_t object.
*
*	p_log
*		[in] Pointer to the log object.
*
*	p_disp
*		[in] Pointer to the OpenSM central Dispatcher.
*
* RETURN VALUES
*	CL_SUCCESS if the MultiPath Record Receive Controller object was
*	initialized successfully.
*
* NOTES
*	Allows calling other MultiPath Record Receive Controller methods.
*
* SEE ALSO
*	MultiPath Record Receive Controller object, osm_pr_rcv_ctrl_construct,
*	osm_mpr_rcv_ctrl_destroy, osm_mpr_rcv_ctrl_is_inited
*********/

/****f* OpenSM: MultiPath Record Receive Controller/osm_mpr_rcv_ctrl_is_inited
* NAME
*	osm_mpr_rcv_ctrl_is_inited
*
* DESCRIPTION
*	Indicates if the object has been initialized with osm_mpr_rcv_ctrl_init.
*
* SYNOPSIS
*/
boolean_t osm_mpr_rcv_ctrl_is_inited(
	IN const osm_mpr_rcv_ctrl_t* const p_ctrl );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to an osm_mpr_rcv_ctrl_t object.
*
* RETURN VALUES
*	TRUE if the object was initialized successfully,
*	FALSE otherwise.
*
* NOTES
*	The osm_mpr_rcv_ctrl_construct or osm_mpr_rcv_ctrl_init must be
*	called before using this function.
*
* SEE ALSO
*	MultiPath Record Receive Controller object, osm_mpr_rcv_ctrl_construct,
*	osm_mpr_rcv_ctrl_init
*********/

END_C_DECLS

#endif	/* _OSM_MPRCTRL_H_ */
